# zealot

Catch N+1 queries in your Django project.

![Static Badge](https://img.shields.io/badge/license-MIT-brightgreen)
![PyPI - Version](https://img.shields.io/pypi/v/zealot?color=lightgrey)

## Features

- Detects N+1s from missing prefetches and from use of `.defer()`/`.only()`
- Friendly error messages like `N+1 detected on User.followers at myapp/views.py:25 in get_user`
- Configurable thresholds
- Allow-list
- Well-tested
- No dependencies

## Acknowledgements

This library draws heavily from jmcarp's [nplusone](https://github.com/jmcarp/nplusone/).
It's not exactly a fork, but not far from it.

## Installation

First:

```
pip install zealot
```

Then, add zealot to your `INSTALLED_APPS` and `MIDDLEWARE`. You probably
don't want to run it in production: I haven't profiled it but it will have a performance
impact.

```python
if DEBUG:
    INSTALLED_APPS.append("zealot")
    MIDDLEWARE.append("zealot.middleware.zealot_middleware")
```

This will detect N+1s that happen in web requests. To catch N+1s in more places,
read on!

### Celery

If you use Celery, you can configure this using [signals](https://docs.celeryq.dev/en/stable/userguide/signals.html):

```python
from celery.signals import task_prerun, task_postrun
from zealot import setup, teardown
from django.conf import settings

@task_prerun.connect()
def setup_zealot(*args, **kwargs):
    setup()

@task_postrun.connect()
def teardown_zealot(*args, **kwargs):
    teardown()
```

### Tests

Django [runs tests with `DEBUG=False`](https://docs.djangoproject.com/en/5.0/topics/testing/overview/#other-test-conditions),
so to run zealot in your tests, you'll first need to ensure it's added to your
`INSTALLED_APPS` and `MIDDLEWARE`. You could do something like:

```python
import sys

TEST = "test" in sys.argv
if DEBUG or TEST:
    INSTALLED_APPS.append("zealot")
    MIDDLEWARE.append("zealot.middleware.zealot_middleware")
```

This will enable zealot in any tests that go through your middleware. If you want to enable
it in _all_ tests, you need to do a bit more work.

If you use pytest, use a fixture in your `conftest.py`:

```python
import pytest
from zealot import zealot_context

@pytest.fixture(scope="function", autouse=True)
def use_zealot():
    with zealot_context():
        yield
```

If you use unittest, add a custom test runner:

```python
# In e.g. `myapp/testing/test_runners.py`
from zealot import setup as zealot_setup, teardown as zealot_teardown
from django.test.runner import DiscoverRunner
from unittest.runner import TextTestResult

class ZealotTestResult(TextTestResult):
    def startTest(self, test):
        zealot_setup()
        return super().startTest(test)

    def addError(self, test, err) -> None:
        zealot_teardown()
        return super().addError(test, err)

    def addFailure(self, test, err) -> None:
        zealot_teardown()
        return super().addFailure(test, err)

    def addSuccess(self, test):
        zealot_teardown()
        return super().addSuccess(test)

class ZealotTestRunner(DiscoverRunner):
    def get_resultclass(self):
        return ZealotTestResult


# And in your settings:
TEST_RUNNER = (
    "myapp.testing.test_runners.ZealotTestRunner"
)
```

### Generic setup

If you also want to detect N+1s in other places not covered here, you can use the `setup` and
`teardown` functions, or the `zealot_context` context manager:

```python
from zealot import setup, teardown, zealot_context


def foo():
    setup()
    try:
        # your code goes here
    finally:
        teardown()


@zealot_context()
def bar():
    # your code goes here


def baz():
    with zealot_context():
        # your code goes here
```

## Configuration

By default, any issues detected by zealot will raise a `ZealotError`. If you'd
rather log any detected N+1s, you can set:

```python
ZEALOT_RAISE = False
```

N+1s will be reported when the same query is executed twice. To configure this
threshold, set the following in your Django settings.

```python
ZEALOT_NPLUSONE_THRESHOLD = 3
```

To handle false positives, you can temporarily disable zealot in parts of your code
using a context manager:

```python
from zealot import zealot_ignore

with zealot_ignore():
    # code in this block will not log/raise zealot errors
```

Finally, if you want to ignore N+1 alerts from a specific model/field globally, you can
add it to your settings:

```python
ZEALOT_ALLOWLIST = [
    {"model": "polls.Question", "field": "options"},

    # you can use fnmatch syntax in the model/field, too
    {"model": "polls.*", "field": "options"},

    # if you don't pass in a field, all N+1s arising from the model will be ignored
    {"model": "polls.Question"},
]
```

## Contributing

1. First, install [uv](https://github.com/astral-sh/uv).
2. Create a virtual env using `uv venv` and activate it with `source .venv/bin/activate`.
3. Run `make install` to install dev dependencies.
4. To run tests, run `make test`.
